# This file is part of versuchung.
#
# versuchung is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# versuchung is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# versuchung.  If not, see <http://www.gnu.org/licenses/>.


from versuchung.types import InputParameter, OutputParameter, Type
import versuchung.archives
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO
import shutil
import csv
import os, stat
import hashlib
import fnmatch

class FilesystemObject(InputParameter, OutputParameter, Type):
    def __init__(self, default_name=""):
        InputParameter.__init__(self)
        OutputParameter.__init__(self)
        Type.__init__(self)
        self.__object_name = default_name
        self.__enclosing_directory = os.path.abspath(os.curdir)
        self.__force_enclosing_directory = False

    def inp_setup_cmdline_parser(self, parser):
        self.inp_parser_add(parser, None, self.__object_name)

    def inp_extract_cmdline_parser(self, opts, args):
        self.__object_name = self.inp_parser_extract(opts, None)

    def inp_metadata(self):
        return {self.name: self.__object_name}

    @property
    def path(self):
        """:return: string -- path to the file/directory"""
        if not self.__force_enclosing_directory:
            if self.parameter_type == "input":
                if self.static_experiment == self.dynamic_experiment:
                    self.__enclosing_directory = self.dynamic_experiment.startup_directory
                else:
                    self.__enclosing_directory = self.static_experiment.base_directory
            elif self.parameter_type == "output":
                assert self.static_experiment == self.dynamic_experiment
                self.__enclosing_directory = self.dynamic_experiment.base_directory
            elif self.static_experiment is not None:
                self.__enclosing_directory = self.static_experiment.base_directory
            else:
                self.__enclosing_directory = os.path.abspath(os.curdir)

        if os.path.isabs(self.__object_name):
            return self.__object_name

        return os.path.join(self.__enclosing_directory, self.__object_name)

    @property
    def basename(self):
        return os.path.basename(self.path)

    @property
    def dirname(self):
        return os.path.dirname(self.path)

    def set_path(self, base_directory, object_name):
        assert base_directory[0] == "/"
        self.__force_enclosing_directory = True
        self.__enclosing_directory = base_directory
        self.__object_name         = object_name

class File(FilesystemObject):
    """Can be used as: **input parameter** and **output parameter**

    The File type represents the content of a single file. Its contents
    can be read and written most easily with the :attr:`value` property.

    Alternatively, the method :meth:`write` appends new content if the
    parameter `append` is set to `True`.

    NB: The content of the file is flushed only after the experiment
    finishes.  Use :meth:`flush` to force writing the buffered data to
    disk before the experiment finishes.
    """

    def __init__(self, default_filename="", binary=False):
        FilesystemObject.__init__(self, default_filename)
        self.__value = None

        self.__binary = binary
        if binary:
            self.__binary_mode = "b"
        else:
            self.__binary_mode = ""


    @property
    def value(self):
        """This attribute can be read and written and represent the
        content of the specified file"""
        if not self.__value:
            try:
                with open(self.original_path, "r" + self.__binary_mode) as fd:
                    self.__value = self.after_read(fd.read())
            except IOError:
                # File couldn't be read
                self.__value = self.after_read("")
        return self.__value

    @value.setter
    def value(self, value):
        self.__value = value

    @property
    def original_path(self):
        return File.path.fget(self)

    def write(self, content, append = False):
        """Similar to the :attr:`value` property. If the parameter
        `append` is `False`, then the property :attr:`value` is reset
        (i.e., overwritten), otherwise the content is appendend"""
        if append:
            self.value += content
        else:
            self.value = content

    def after_experiment_run(self, parameter_type):
        FilesystemObject.after_experiment_run(self, parameter_type)
        assert parameter_type in ["input", "output"]
        if parameter_type == "output":
            self.flush()

    def flush(self):
        """Flush the cached content of the file to disk"""
        if self.__value == None:
            return
        with open(self.original_path, "w" + self.__binary_mode + "+") as fd:
            v = self.before_write(self.value)
            if v is None:
                v = ""
            fd.write(v)

    def copy_contents(self, filename):
        """Read the given file and replace the current .value with the
        files content.

        Flushes automatically afterwards."""
        with open(filename) as fd:
            self.value = self.after_read(fd.read())
        self.flush()

    def make_executable(self):
        """makes a file exectuable (chmod +x $file)"""
        st = os.stat(self.original_path)
        os.chmod(self.original_path, st.st_mode | stat.S_IEXEC)

    def after_read(self, value):
        """To provide filtering of file contents in subclasses, overrwrite this method.
        It is gets the file content as a string and returns the value()"""
        return value

    def before_write(self, value):
        """To provide filtering of file contents in subclasses, overrwrite this method.
        This method gets the value() and returns a string, when the file is written to disk"""
        return value

class Executable(File):
    """Can be used as: **input parameter**

    An executable is a :class:`versuchung.files.File` that only
    references an executable. It checksums the executable and puts the
    checksum into the metadata. The executable is never changed.

    """
    def __init__(self, default_filename):
        File.__init__(self, default_filename)

    @property
    def value(self):
        raise NotImplemented

    @value.setter
    def value(self, value):
        raise NotImplementedError

    def write(self, content, append = False):
        raise NotImplementedError

    def after_experiment_run(self, parameter_type):
        pass

    def flush(self):
        raise NotImplementedError

    def copy_contents(self, filename):
        raise NotImplementedError

    def make_executable(self):
        raise NotImplementedError

    def inp_metadata(self):
        return {self.name + "-md5": hashlib.md5(open(self.path, "rb").read()).hexdigest()}

    def execute(self, cmdline, *args):
        """Does start the executable with meth:`versuchung.execute.shell` and
        args, which is of type list, as arguments."""
        from versuchung.execute import shell

        shell(self.path + " " + cmdline, *args)

class Directory_op_with:
    def __init__(self):
        self.__olddir = []
    def __enter__(self):
        self.__olddir.append(os.path.abspath(os.curdir))
        os.chdir(self.path)
        return self.path
    def __exit__(self, *excinfo):
        path = self.__olddir[-1]
        del self.__olddir[-1]
        os.chdir(path)


class Directory(FilesystemObject, Directory_op_with):
    """Can be used as: **input parameter** and **output parameter**

    Represents the contents of directory. The filename_filter is a
    glob/fnmatch expression to filter the directories content and to
    ensure that no file is generated that does not fit this pattern.
    An useful example of this is an output Directory that matches only
    *.log files and is directly located in the result directory:

       outputs = {
           "logs": Directory(".", filename_filter="*.log")
       }

    It can also be used with the **with**-keyword to change the
    current working directory temporarily to this directory::

       with directory as dir:
          # Do something with adjusted current working directory
          print os.curdir

    """

    def __init__(self, default_filename="", filename_filter="*"):
        FilesystemObject.__init__(self, default_filename)
        Directory_op_with.__init__(self)
        self.filename_filter = filename_filter
        self.__value = None
        self.__new_files = []

    def __ensure_dir_exists(self):
        if not os.path.exists(self.path):
            os.mkdir(self.path)

    @property
    def value(self):
        """:return: list -- directories and files in given directory"""
        if not self.__value:
            self.__value = os.listdir(self.path)
            self.__value = [x for x in self.__value
                            if fnmatch.fnmatch(x, self.filename_filter)]
        return self.__value

    def __iter__(self):
        for name in self.value:
            p = os.path.join(self.path, name)
            if name in self.subobjects:
                yield self.subobjects[name]
                continue

            if os.path.isdir(p):
                d = Directory(name)
                d.set_path(self.path, p)
                self.subobjects[name] = d
                yield d
            else:
                if p.endswith(".gz"):
                    f = versuchung.archives.GzipFile(name)
                else:
                    f = File(name)
                f.set_path(self.path, p)
                self.subobjects[name] = f
                yield f

    def before_experiment_run(self, parameter_type):
        FilesystemObject.before_experiment_run(self, parameter_type)
        if parameter_type == "output":
            self.__ensure_dir_exists()

    def new_file(self, name, compressed=False):
        """Generate a new :class:`~versuchung.files.File` in the
        directory. It will be flushed automatically if the experiment
        is over."""
        if not fnmatch.fnmatch(name, self.filename_filter):
            raise RuntimeError("Filename {} does not match filter {}".\
                                 format(name, self.filename_filter))
        self.__ensure_dir_exists()
        if compressed:
            f = versuchung.archives.GzipFile(name)
        else:
            f = File(name)
        f.set_path(self.path, name)
        f.value = ""
        self.subobjects[name] = f
        return f

    def new_directory(self, name):
        """Generate a new :class:`~versuchung.files.Directory` in the
        directory. The directory <name> must not be present before"""
        if not fnmatch.fnmatch(name, self.filename_filter):
            raise RuntimeError("Filename {} does not match filter {}".\
                                 format(name, self.filename_filter))
        self.__ensure_dir_exists()
        f = Directory(name)
        f.set_path(self.path, name)
        os.mkdir(f.path)
        self.subobjects[name] = f
        return f


    def mirror_directory(self, path, include_closure = None):
        """Copies the contents of the given directory to this
        directory.

        The include closure is a function, which checks for every
        (absolute) path in the origin directory, if it is mirrored. If
        it is None, all files are included."""

        self.__ensure_dir_exists()

        if not include_closure:
            include_closure = lambda arg: True

        if not os.path.exists(path) and os.path.isdir(path):
            raise RuntimeError("Argument is no directory")

        path = os.path.abspath(path)

        for root, dirs, files in os.walk(path):
            root = root[len(path)+1:]
            for d in dirs:
                src = os.path.join(path, root, d)
                if not include_closure(src):
                    continue
                dst = os.path.join(self.path, root, d)
                if not os.path.isdir(dst):
                    os.mkdir(dst)
            for f in files:
                src = os.path.join(path, root, f)
                if not include_closure(src):
                    continue
                dst = os.path.join(self.path, root, f)
                shutil.copyfile(src,dst)

        self.__value = None

class CSV_File(File):
    """Can be used as: **input parameter** and **output parameter**

    It is a normal :class:`~versuchung.files.File` but the content of the file is
    interpreted as a csv file. It is parsed before the value is
    exposed to the user. And formatted before the content is written
    to disk.

    Internally the :mod:`csv` is used, so all arguments to
    ``csv.reader`` and ``csv.writer`` can be given in *csv_args*."""

    value = File.value
    """Other than a normal CSV_File the value of a CSV_File is a list
    of lists, which represents the structure of the csv file. This
    value can be manipulated by the user.

    >>> CSV_File("csv_output").value
    [["1", "2", "3"]]"""

    def __init__(self, default_filename = "", **csv_args):
        File.__init__(self, default_filename)
        self.csv_args = csv_args

    def after_read(self, value):
        fd = StringIO(value)
        reader = csv.reader(fd, self.csv_args)
        return list(reader)
    def before_write(self, value):
        fd = StringIO()
        writer = csv.writer(fd, self.csv_args)
        writer.writerows(value)
        return fd.getvalue()

    def write(self):
        raise NotImplemented

    def append(self, row):
        """Append a row to the csv file

        It is just a shorthand for

        >>> csv_file.value.append([1,2,3])

        :param row: row to append
        :type row: list."""
        if type(row) != list:
            raise TypeError("list of values required")
        self.value.append(row)



# This file is part of versuchung.
#
# versuchung is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# versuchung is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# versuchung.  If not, see <http://www.gnu.org/licenses/>.

from versuchung.experiment import Experiment

class JupyterExperiment(Experiment):
      def __init__(self, title, *args, **kwargs):
            Experiment.__init__(self, *args, title=title, **kwargs)

      def execute(self):
            raise RuntimeError("Please use begin()/end()")

      def begin(self, args=[], globals=None):
            if globals is None:
                  raise RuntimeError("Please specify globals=globals()")
            self._globals = globals
            args = args + self._globals.get('versuchung_args', [])
            self.execute_setup(args)
            out = self._globals.get('versuchung_path')
            if out:
                  with open(out, "w+") as fd:
                        fd.write(self.path)

      def end(self):
            self.execute_teardown()




# This file is part of versuchung.
#
# versuchung is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# versuchung is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# versuchung.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

from versuchung.types import Type, InputParameter
from versuchung.files import Directory, Directory_op_with, File
from versuchung.execute import shell
import logging
import os
import sys
import gzip
import re
from subprocess import PIPE
try:
    from StringIO import StringIO as BytesIO
except Exception:
    from io import BytesIO

class TarArchive(Type, InputParameter, Directory_op_with):
    """Can be used as: **input parameter**

    The archive will be extracted to a temporary directory. It will be
    removed after the experiment is over.

    ``clone_url`` can either be a :class:`string` or any object that
    has a ``.path`` attribute (like
    e.g. :class:`~versuchung.filesystems.File`). Of course the
    referenced file must be a single file.

    This parameter can be used as argument to the with keyword, to
    change to the temporary directory::

        with self.inputs.tar_archive as path:
            # Here we have path == os.path.abspath(os.curdir)
            # Do something in the extracted copy
            print path
    """
    def __init__(self, filename = None):
        """The default_filename is either a string to a file. Or a
        object with a path attribute (e.g. a :class:`~versuchung.files.File`)"""
        Type.__init__(self)
        InputParameter.__init__(self)
        Directory_op_with.__init__(self)

        self.__filename = filename
        self.__value = None

    def inp_setup_cmdline_parser(self, parser):
        self.inp_parser_add(parser, None, self.__filename)

    def inp_extract_cmdline_parser(self, opts, args):
        self.__filename = self.inp_parser_extract(opts, None)

    def before_experiment_run(self, parameter_type):
        if parameter_type == "input" and "path" in dir(self.__filename):
            self.subobjects["filename"] = self.__filename
            Type.before_experiment_run(self, parameter_type)
            self.__filename = self.__filename.path
        else:
            Type.before_experiment_run(self, parameter_type)

        self.__filename = os.path.abspath(self.__filename)

    def inp_metadata(self):
        return {self.name: self.__filename}

    def __setup_value(self):
        if "path" in dir(self.__filename):
            self.subobjects["filename"] = self.__filename
            self.__filename = self.__filename.path

        self.__filename = os.path.abspath(self.__filename)
        fn = self.__filename

        extract_mode = ""
        if "tar.gz" in fn or "tgz" in fn:
            extract_mode = "z"
        if "tar.bz2" in fn or "bzip2" in fn:
            extract_mode = "j"
        if "tar.xz" in fn or "txz" in fn:
            extract_mode = "J"

        with self.tmp_directory as d:
            try:
                os.mkdir(self.name)
            except OSError:
                # ignore errors if the directory should already exist for some reason
                pass
            with Directory(self.name) as d2:
                dirname = os.path.abspath(".")
                (out, ret) = shell("tar %sxvf %s", extract_mode, fn)
                if ret != 0:
                    raise RuntimeError("Extracting of %s failed" % fn)

                cd = None
                for line in out:
                    if (cd == None or len(line) < len(cd)) and line.endswith("/"):
                        cd = line
                if cd and all([x.startswith(cd) for x in out]):
                    dirname = cd
                return Directory(os.path.abspath(dirname))

    @property
    def value(self):
        """Return a :class:`versuchung.files.Directory` instance to the extracted
        tar archive. If it contains only one directory the instance
        will point there. Otherwise it will point to a directory
        containing the contents of the archive"""
        if not self.__value:
            self.__value = self.__setup_value()
        return self.__value

    @property
    def path(self):
        """Return the string to the extract directory (same as .value.path)"""
        return self.value.path


class GitArchive(InputParameter, Type, Directory_op_with):
    """Can be used as: **input parameter**

    The git repository given in ``clone_url`` will be cloned to a
    temporary directory. It will be removed after the experiment is
    over. If ``shallow == True`` Only the files and not the .git is
    copied (cloned). This is especially useful for large git
    repositories like the Linux kernel tree.

    ``clone_url`` can either be a :class:`string` or any object that
    has a ``.path`` attribute (like e.g. :class:`TarArchive`). Of
    course the refenced path must be a directory.

    This parameter can be used as argument to the with keyword, to
    change to the temporary directory::

        with self.inputs.git_archive as path:
            # Here we have path == os.path.abspath(os.curdir)
            # Do something in the extracted copy
            print path
    """

    def __init__(self, clone_url = None, ref = "refs/heads/master", shallow=False,
                 branches=None, tags=None):
        """clone_url: where to the git archive from
              This might either be a string or a object with a path attribute
           ref: which git reference to checkout
           shallow: do a shallow copy (using git-archive).
           branches: Also fetch other branches. Use branches=True for all branches and branches=REGEX for a filtered view.
           tags:     Also fetch other tags. Use tags=True for all tags and tags=REGEX for a filtered view.


           The git archive will be cloned to self.name (which is the
           key in the input parameters dict)"""
        Type.__init__(self)
        InputParameter.__init__(self)
        Directory_op_with.__init__(self)

        self.__clone_url = clone_url
        self.__ref = ref
        self.__shallow = shallow
        self.__value = None
        self.__hash = None

        # Include branches and tags into the metadata-hash
        self.__filter_refs = {"branches": branches, "tags": tags}

        if (branches or tags ) and shallow:
             raise RuntimeError("Shallow clone and branch/tag checkout is not supported.")

    def inp_setup_cmdline_parser(self, parser):
        self.inp_parser_add(parser, "clone-url", self.__clone_url)
        self.inp_parser_add(parser, "ref", self.__ref)


    def inp_extract_cmdline_parser(self, opts, args):
        self.__clone_url = self.inp_parser_extract(opts, "clone-url")
        self.__ref = self.inp_parser_extract(opts, "ref")

    def before_experiment_run(self, parameter_type):
        if parameter_type == "input" and "path" in dir(self.__clone_url):
            self.subobjects["clone-url"] = self.__clone_url
            Type.before_experiment_run(self, parameter_type)
            self.__clone_url = self.__clone_url.path
        else:
            Type.before_experiment_run(self, parameter_type)

    def __references(self, prefix_filter=None, regex_filter=None):
        (lines, ret) = shell("git ls-remote %s 'refs/*'", self.__clone_url, stderr=sys.stderr)

        if ret != 0 or lines == 0:
            print("\n".join(lines))
            raise RuntimeError("Could not list references in repository")

        ret = {}
        for line in lines:
            commit_hash, refname = line.strip().split("\t", 1)

            # Apply regex and prefix filter
            if prefix_filter:
                if not refname.startswith(prefix_filter):
                    continue
                # I have no idea where these names come from
                if refname.startswith("refs/tags/") and refname.endswith("^{}"):
                    continue
                # Shorten the Prefix
                refname = refname[len(prefix_filter):]

            if regex_filter and not re.match(regex_filter, refname):
                continue

            ret[refname] = commit_hash

        return ret

    def references(self, regex_filter=None):
        """Inspect reference list of the repository. This reference list
           includes all branches, tags, and whatsoever. All reference names are
           fully qualified (refs/tags/*, refs/heads*).

           If the ``regex_filter`` is given, only reference names that
           match the regex are returned.

           @returns a dictionary that maps references to commit hashes

        """
        return self.__references(regex_filter=regex_filter)

    def tags(self):
        """Like references, but returns a list of tags in the repository. The
           keys are transformed (refs/tags/$X -> $X).
           The GitArchive(tags=...) filter applies!


           @returns dictionary of tag name to commit hashes

        """
        regex_filter = None
        arg  = self.__filter_refs["tags"]
        if arg and arg is not True:
            regex_filter = arg
        return self.__references(prefix_filter="refs/tags/", regex_filter=regex_filter)

    def branches(self):
        """Like references, but returns a list of branches/heads in the
           repository. The keys are transformed (refs/heads/$X -> $X).
           The GitArchive(branches=...) filter applies!

           @returns dictionary of tag name to commit hashes

        """
        regex_filter = None
        arg  = self.__filter_refs["branches"]
        if arg and arg is not True:
            regex_filter = arg
        return self.__references(prefix_filter="refs/heads/", regex_filter=regex_filter)


    def checkout(self, branch=None, tag=None):
        if branch:
            visible_branches = self.__metadata.get("branches", {})
            if branch not in visible_branches:
                raise RuntimeError("Branch `%s' is not visible, please parametrize GitArchive(branches=...) correctly" % branch)
            self.__ref = "refs/heads/" + branch
            self.__hash = visible_branches[branch]
        if tag:
            visible_tags = self.__metadata.get("tags", {})
            if tag not in visible_tags:
                raise RuntimeError("Tag `%s' is not visible, please parametrize GitArchive(tags=...) correctly" % tag)
            self.__ref =  "refs/tags/" + tag
            self.__hash = visible_tags[tag]

        if not self.__ref:
            raise RuntimeError("GitArchive.checkout() requires branch or tag parameter")

        cmd = "cd '%s' && git checkout %s"
        (lines, ret) = shell(cmd, self.value.path, self.__ref, stderr=sys.stderr)
        if ret != 0:
            print("\n".join(lines))
            raise RuntimeError("GitArchive.checkout(%s) failed" % self.__ref)

        return (self.__ref, self.__hash)


    def checkout_hash(self):
        """Return the hash of the HEAD commit hash as string"""
        if not self.__hash:
            cmd = "git ls-remote %s %s" % (self.__clone_url,
                                           self.__ref)

            (lines, ret) = shell(cmd, stderr=sys.stderr)
            if ret != 0 or lines == 0:
                print("\n".join(lines))
                sys.exit(-1)

            self.__hash = lines[0].split("\t")[0]
            if self.__hash == "":
                self.__hash = self.__ref

        return self.__hash

    def checkout_ref(self):
        """Return git ref which was checked out"""
        return self.__ref

    def checkout_url(self):
        """Return git url which was checked out"""
        return self.__clone_url

    def inp_metadata(self):
        ret = {"clone-url": str(self.__clone_url),
               "ref": self.__ref,
               "hash": self.checkout_hash()}

        # If the user wants to checkout different branches, these have
        # to influence the metadata hash
        for ref, getter in [("branches", self.branches),
                            ("tags",     self.tags)]:
            if self.__filter_refs[ref]:
                ret[ref] = getter()

        self.__metadata = ret

        return {self.name + "-" + k: v for (k,v) in ret.items()}

    def __setup_value(self):
        if "path" in dir(self.__clone_url):
            self.subobjects["clone-url"] = self.__clone_url
            self.__clone_url = self.__clone_url.path

        logging.info("copying git archive %s", self.__clone_url)
        with self.tmp_directory as d:
            os.mkdir(self.name)
            if self.__shallow:
                cmd = "cd '%s' && git archive --format=tar --remote=%s %s | tar x"
                args = (self.name,
                        self.__clone_url,
                        self.__ref)
            else:
                cmd = "git clone %s %s"
                args = (self.__clone_url, self.name)

            (lines, ret) = shell(cmd, *args, stderr=sys.stderr)

            if ret != 0:
                print("\n".join(lines))
                sys.exit(-1)

            if not self.__shallow:
                cmd = "cd %s && git gc && git fetch %s %s && git checkout FETCH_HEAD"
                args = (self.name, self.__clone_url, self.__ref)
                (lines, ret) = shell(cmd, *args, stderr=sys.stderr)

                if ret != 0:
                    print("\n".join(lines))
                    sys.exit(-1)

                # Fetch all visible branches and tags
                for branch in self.__metadata.get("branches", {}):
                    cmd = "cd %s && git fetch %s refs/heads/%s && git update-ref refs/heads/%s FETCH_HEAD"
                    shell(cmd, self.name, self.__clone_url, branch, branch, stderr=sys.stderr)
                for tag in self.__metadata.get("tags", {}):
                    cmd = "cd %s && git fetch %s refs/tags/%s && git update-ref refs/tags/%s FETCH_HEAD"
                    shell(cmd, self.name, self.__clone_url, tag, tag, stderr=sys.stderr)

            return Directory(os.path.abspath(self.name))

    @property
    def value(self):
        """Return a :class:`versuchung.files.Directory` instance to the cloned git directory"""
        if not self.__value:
            self.__value = self.__setup_value()
        return self.__value

    @property
    def path(self):
        """Return the string to the extract directory (same as .value.path)"""
        return self.value.path


class GzipFile(File):
    def __init__(self, default_filename=""):
        File.__init__(self, default_filename, binary=True)

    @property
    def path(self):
        """Decompress file into the temporary directory and return path to this location"""
        assert self.tmp_directory is not None, \
            "Can gunzip file only as part of an active experiment"

        path = File.path.fget(self)
        base = os.path.basename(path.rstrip(".gz"))
        filename = os.path.join(self.tmp_directory.path,
                                self.name + "_" + base)

        if not os.path.exists(filename):
            shell("gunzip < %s > %s", path, filename)

        return filename

    def after_read(self, value):
        x = BytesIO(value)
        fd = gzip.GzipFile(fileobj=x)
        return fd.read().decode()

    def before_write(self, value):
        x = BytesIO()
        fd = gzip.GzipFile(fileobj=x, mode="w")
        fd.write(value.encode())
        fd.close()
        return x.getvalue()



#!/usr/bin/python3
import os
import sys
import shutil
import datetime
import logging

def transitive_hull(experiments, from_metadata=False):
    good = set()
    roots = list(experiments)
    while roots:
        item = roots.pop()
        item = os.path.relpath(item)
        if not os.path.exists(item):
            continue
        if not os.path.exists(os.path.join(item, 'metadata')):
            continue
        if item in good:
            continue
        good.add(item)
        if os.path.islink(item):
            roots.append(os.path.realpath(item))

        if not from_metadata:
            continue

        m = eval(open(os.path.join(item, "metadata")).read())
        for value in m.values():
            if type(value) is str:
                roots.append(value)
    return good

def print_list(items, verbose = False):
    symlinks = [x for x in items if os.path.islink(x)]
    deref    = [os.path.realpath(x) for x in symlinks]
    items =    [x for x in items if os.path.abspath(x) not in deref]

    max_name = max(*[len(x) for x in items])

    lines = []
    total_runtime = None
    for item in sorted(items, key = lambda x: (x not in symlinks, x)):
        metadata_path = os.path.join(item, "metadata")
        if not os.path.exists(metadata_path):
            continue

        if not verbose:
            print(item)
            continue

        # With verbosity
        with open(metadata_path) as fd:
            try:
                metadata = eval(fd.read())
            except:

                lines.append([item, 'broken'])
                continue
        if 'date-end' not in metadata:
            lines.append([item, 'incomplete',
                          "(begin: %s)"%metadata['date-start']])
            continue
        start = datetime.datetime.strptime(metadata['date-start'], "%Y-%m-%d %H:%M:%S.%f")
        end   = datetime.datetime.strptime(metadata['date-end'], "%Y-%m-%d %H:%M:%S.%f")
        delta = end-start
        delta = datetime.timedelta(delta.days, delta.seconds) # Drop microseconds

        if total_runtime:
            total_runtime += delta
        else:
            total_runtime = delta

        lines.append([item, 'ok', str(delta)])
    align = 0
    for line in lines:
        align = max(align, len(line[0]))
    for line in lines:
        print("{0:<{1}}".format(line[0], align), *line[1:])

    print("Total Runtime: ", total_runtime)

def main(argv):
    p = os.path.realpath(__file__)
    p = os.path.abspath(p)
    p = os.path.dirname(os.path.dirname(p))
    sys.path.append(p)
    if len(argv) < 1:
        print(f"usage: {sys.argv[0]} <CMD | ipynb>")
        print("")
        print("\tls     - List all experiment results in this directory")
        print("\tgc     - Results that are not referenced by symlink or from metadata")
        print("\tipynb  - Run Jupyter Notebook as experiment")
        sys.exit(1)

    if argv[0] == "gc":
        if len(sys.argv) > 1 and sys.argv[1] == '-v':
            del sys.argv[1]
            verbose = True
        else:
            verbose = False
        roots = sys.argv[1:]
        if not roots:
            roots = [x for x in os.listdir(".") if os.path.islink(x)]

        good = transitive_hull(roots, from_metadata=True)
        items = set(os.listdir(".")) - good

        print_list(items, verbose=verbose)

    elif argv[0] == "ls":
        print_list(os.listdir("."), verbose=True)

    elif argv[0].endswith(".ipynb"):
        import papermill as pm
        from tempfile import NamedTemporaryFile
        from nbconvert import HTMLExporter

        output = NamedTemporaryFile(suffix=".ipynb")
        path = NamedTemporaryFile(mode="r+")
        nb = pm.execute_notebook(
            input_path=argv[0],
            output_path=output.name,
            cwd=os.path.dirname(os.path.join(".", argv[0])),
            report_mode=True,
            log_output=True,
            stdout_file=sys.stdout,
            stderr_file=sys.stderr,
            progress_bar=False,
            parameters=dict(versuchung_args=argv[1:],
                            versuchung_path=path.name))

        # Copy Output Jupyter Notebook to directory
        experiment_dir = path.read()
        if not experiment_dir:
            raise RuntimeError("Notebook does not experiment path. Did you call experiment.begin()")
        dst = os.path.join(experiment_dir, os.path.basename(argv[0]))
        with open(dst, "wb+") as ipynb:
            ipynb.write(output.read())

        html_exporter = HTMLExporter(template_name = 'classic')
        (body, resources) = html_exporter.from_notebook_node(nb)
        with open(dst +".html", "w+") as html:
            html.write(body)
        return experiment_dir

    elif argv[1] == "cp":
        if len(sys.argv) < 3:
            print("versuchung gc <dst 1> [<dst 2>...] <dst>")
            sys.exit(-1)
        roots = sys.argv[2:-1]
        target = sys.argv[-1]
        if os.path.exists("%s/metadata" % target):
            print("ERROR: target is an experiment instance. Forgot dst?")
            sys.exit(-1)
        srcs = transitive_hull(roots)
        for src in sorted(srcs):
            dst = os.path.join(target, src)
            if os.path.islink(src):
                if os.path.islink(dst):
                    os.unlink(dst)
                value = os.readlink(src)
                os.symlink(value, dst)
            else:
                if os.path.exists(dst):
                    print("WARN: %s already exists" % src)
                    continue
                else:
                    shutil.copytree(src, dst)
            print("%s -> %s/"% (src, target))
    else:
        main([])

if __name__ == "__main__":
    main(sys.argv[1:])
